/*-----------------------------------------------------------------------------

 2002, Jeff M Soft und Hardware GmbH, All Rights Reserved

-----------------------------------------------------------------------------*/

#include "Module.h"
#include "SEMod_struct.h"
#include "SEPin.h"
#include <assert.h>
#include <math.h>

#include <float.h>
#include <windows.h>
#include <stdio.h>

// define some constants to make referencing in/outs clearer
#define PN_INPUT1 0
#define PN_INPUT2 1
#define PN_OUTPUT1 2

// handy function to fix denormals.
static const float anti_denormal = 1e-18f;
inline void kill_denormal(float &p_sample ) { p_sample += anti_denormal;p_sample -= anti_denormal;};

// The most important part, processing the audio
void Module::sub_process(long buffer_offset, long sampleFrames )
{
	// assign some pointers to your in/output buffers. usually blocks (array) of 96 samples
    float *out1 = buffer_offset + output1_buffer;
	float *in1 = buffer_offset + input1_buffer;
	float *in2 = buffer_offset + input2_buffer;

	// float sr = getSampleRate();

	const float four = 4.f;
	const float one = 1.f;
	float zero = 0.f;
	kill_denormal(zero);

	int p4i;
	float p4f;

	float freq;
	float fatness;
	int ilookup;
	float flookup;

	float ph = phase;
	kill_denormal(ph);
	float v;

	/*
	_set_SSE2_enable (
		1
	);
	_control87 (
		_EM_INVALID | _EM_DENORMAL | _EM_ZERODIVIDE | _EM_OVERFLOW | _EM_UNDERFLOW | _EM_INEXACT,
		_MCW_EM
	);
	*/

	short control_word;
	short control_word2;

	__asm
	{
		fstcw	control_word		// store fpu control word
		mov 	dx, word ptr [control_word]
 		or		dx,0x0400			// round towards -oo
		mov		control_word2, dx
		fldcw	control_word2		// load modfied control word
	}
/*
	bool isLFO = false;
	float *table;
	if(*in1 < 0.f) {
		isLFO = true;
		table = n_table;
	} else {
		table = p_table;
	}
*/
	/*
	flookup *= FATOSC_TABLESIZE_MINUS_1;
	ilookup = (int) flookup;
	flookup -= ilookup;
	if(ilookup < 0) ilookup = 0;
	if(ilookup > FATOSC_TABLESIZE_MINUS_1_INT) ilookup = FATOSC_TABLESIZE_MINUS_1_INT;
	// freq = table[ilookup] * (1-flookup) + table[ilookup+1] * flookup;
	freq = table[ilookup] + (table[ilookup+1]-table[ilookup]) * flookup;
	*/
	const float TSM1 = 1024.f; //  FATOSC_TABLESIZE_MINUS_1;
	const int TSM1i = 1024; // FATOSC_TABLESIZE_MINUS_1_INT;
	// freq = table[ilookup] + (table[ilookup+1]-table[ilookup]) * flookup;

	while(--sampleFrames >= 0)
	{
		/*
		// *** lookup the fatness
		flookup = *in2++ * FATOSC_TABLESIZE_MINUS_1;
		ilookup = (int) flookup;
		flookup -= ilookup;
		if(ilookup < 0) ilookup = 0;
		if(ilookup > FATOSC_TABLESIZE_MINUS_1_INT) ilookup = FATOSC_TABLESIZE_MINUS_1_INT;
		// fatness = f_table[ilookup] * (1-flookup) + f_table[ilookup+1] * flookup;
		fatness = f_table[ilookup] + (f_table[ilookup+1]-f_table[ilookup]) * flookup;
		*/
		/*
		float f_add, f_sub, f_range;
		fatness = *in2++;
		if(fatness <= 0.2f) {
			f_add = 0.0f;
			f_sub = 0.0f;
			f_range = 0.2f/0.25f;
		} else if(fatness <= 0.5f) {
			f_add = 0.25f;
			f_sub = 0.2f;
			f_range = 0.3f/0.75f;
		} else if(fatness <= 0.8f) {
			f_add = 1.0f;
			f_sub = 0.5f;
			f_range = 0.3f/2.0f;
		} else {
			f_add = 3.0f;
			f_sub = 0.8f;
			f_range = 0.2f/97.0f;
		}
		fatness = f_add + ( (fatness - f_sub) / f_range );
		*/
/*
		float afreq = *in1++;
		if(isLFO) {
			afreq += 1.f;
		}

		typedef float (Module::*TypeAPtr)(float, float);
		TypeAPtr _ohapow = &Module::ohapow;
*/

		__asm
		{

		// *** lookup the fatness

			mov     eax, dword ptr [in2]		// st(0) = *in2
			fld		dword ptr [eax]
			add		eax,4						// in2++
			mov		dword ptr [in2],eax
			fmul	dword ptr TSM1				// flookup * TABLESIZE
			fist    dword ptr [ilookup]			// ilookup = (int) flookup
			fisub	dword ptr [ilookup]			// flookup -= ilookup
			mov		edx, dword ptr [ilookup]	// edx = ilookup

			// check table index range

			cmp		edx, 0

			jge		f_not_too_small
			mov     edx, 0						// edx < 0 -> 0
			fstp	st(0)
			fldz
			jmp		f_in_range
f_not_too_small:
			cmp		edx, 1024
			jl		f_in_range
			mov     edx, 1024					// edx > 1024 -> 1024
			fstp	st(0)
			fldz
f_in_range:
			mov		ecx, dword ptr [this]
			mov		ecx, dword ptr[ecx]this.f_table
			shl		edx, 2						// edx *= 4
			lea		esi,[ecx+edx]				// esi = & (table[ilookup] )
			fld		dword ptr [esi]				// table[0]
			fld		dword ptr [esi+4]			// table[1]
			fsub	st(0),st(1)					// table[1] - table[0]
			fmulp	st(2),st(0)					// * flookup
			fadd								// + table[0]
			fstp	fatness

			// *** do the math

			fld		dword ptr ph
			fist    dword ptr [p4i]
			fisub   dword ptr [p4i]
			fstp	p4f

			mov		edx, dword ptr [p4i]
			mov		ecx, dword ptr [this]
			mov		ecx, dword ptr[ecx]this.tx
			lea		esi,[ecx+edx*4]		// esi = & (tx[p4i] )

			//fld		dword ptr [esi]
			//fld		dword ptr [esi+16]
			fld		dword ptr [esi+32]
			fld		dword ptr [esi+48]

//			fmul	dword ptr [esi+48]
//			fadd	dword ptr [esi+32]

//			fld		dword ptr [esi+48]  // tx[p4i+12]
//			fmul
//			fld		dword ptr [esi+32]  // tx[p4i+8]
//			fadd

//			push	esi
//			fstp	p4f
//			fld		dword ptr [esi+16]  // tx[p4i+4]
//			fstp	tb
//			fld		dword ptr [esi]     // tx[p41]
//			fstp	ta

			fld		dword ptr p4f
			fmul
			fadd
			fstp	p4f

		/*
			mov		eax, dword ptr [fatness]
			push	eax
			mov		eax, dword ptr [p4f]
			push	eax

			// call	powf

			lea		ecx, this
			call	[_ohapow]

			pop		ebx
			pop		ebx
			
			// fld		dword ptr [eax]
		*/

		}


/*
		flookup = *in2++ * TSM1;
		ilookup = (int) flookup;
		flookup -= ilookup;
		if(ilookup < 0) { ilookup = 0; flookup = 0.0f; }
		if(ilookup > 1024) { ilookup = 1024; flookup = 0.0f; }
		fatness = f_table[ilookup] + (f_table[ilookup+1]-f_table[ilookup]) * flookup;

		p4f = phase;
		p4i = (int) p4f;
		p4f -= p4i;
*/
		// v = tx[p4i] + tx[p4i+4] * ohapow(tx[p4i+8] + tx[p4i+12]*p4f, fatness);

		// v = tx[p4i] + tx[p4i+4] * pow(tx[p4i+8] + tx[p4i+12]*p4f, fatness);
		v = tx[p4i] + tx[p4i+4] * ohapow(p4f, fatness);
		kill_denormal(v);

		// *out1++ = v;

		__asm {

			/*
			
			mov		edx, dword ptr [p4i]
			mov		ecx, dword ptr [this]
			mov		ecx, dword ptr[ecx]this.tx
			lea		esi,[ecx+edx*4]		// esi = & (tx[p4i] )

			fld		dword ptr [esi]
			fld		dword ptr [esi+16]

			fmul
			fadd

			*/

			fld		dword ptr v
			mov     eax,dword ptr [out1]
			fstp    dword ptr [eax]
			add     eax,4
			mov     dword ptr [out1],eax

		}
/*
		freq = *in1++;
		//if(isLFO) {
		//	freq += 1.0f;
		//}

		flookup = freq * TSM1;
		ilookup = (int) flookup;
		flookup -= ilookup;
		if(ilookup < 0) { ilookup = 0; flookup = 0.0f; }
		if(ilookup > 1024) { ilookup = 1024; flookup = 0.0f; }
		freq = p_table[ilookup] + (p_table[ilookup+1]-p_table[ilookup]) * flookup;
		kill_denormal(freq);
		phase = phase + freq;
		kill_denormal(phase);
		if(phase > 4.0f) phase = 0.0f;
		kill_denormal(phase);
*/
		__asm {

		// *** lookup the frequency (phase increment)


			mov     eax, dword ptr [in1]		// st(0) = *in1
			fld		dword ptr [eax]
			add		eax,4						// in1++
			mov		dword ptr [in1],eax
		
			fcom	zero
			fstsw	ax
			fwait
			sahf
			jb		isLFO
		// *** use p_table
			mov		ecx, dword ptr [this]
			mov		ecx, dword ptr[ecx]this.p_table
			jmp		do_freq_lookup
		isLFO:
		// *** use n_table
			mov		ecx, dword ptr [this]
			mov		ecx, dword ptr[ecx]this.n_table
			fadd	dword ptr one

			// fld		dword ptr freq
		do_freq_lookup:
			fmul	dword ptr TSM1
			fist    dword ptr [ilookup]
			fisub	dword ptr [ilookup]
			mov		edx, dword ptr [ilookup]

			// check table index range

			cmp		edx, 0

			jge		h_not_too_small
			mov     edx, 0
			fstp	st(0)
			fldz
			jmp		h_in_range
	h_not_too_small:
			cmp		edx, 1024
			jl		h_in_range
			mov     edx, 1024
			fstp	st(0)
			fldz
	h_in_range:
			shl		edx, 2						// edx *= 4
			lea		esi,[ecx+edx]		// esi = & (table[int_idx] )
			fld		dword ptr [esi]     // table[0]
			fld		dword ptr [esi+4]   // table[1]
			fsub	st(0),st(1)			// table[1] - table[0]
			fmulp	st(2),st(0)			// * idx_frac
			fadd						// + table[0]
			fstp	freq

		// *** increase the phase

			fld		dword ptr ph
			fadd	dword ptr freq

			fcom	four
			fstsw	ax
			fwait
			sahf
			jna		done
		// *** reset phase
			fstp	st(0)
			fldz
		done:
			fstp	ph

		}

	}
	__asm
	{
		fldcw	control_word
	}
	phase = ph;
}

float Module::ohapow(float a, float b) {
	float lshift;
	float logA, logD, exp1, expA, expD;
	int logI, expI;
	float l, lb;

	if(a < 0.000001f) return 0.0f;
	// *** we should not get this, but...
	if(a > 1.0f) a -= 1.0f;
	lshift = 0.0f;

	while(a*16.0f < 1.0f) {
		a *= 16.0f;
		lshift += log16;
	}

	logA = a*((float)logTSIZE-2.0f);
	logI = (int) logA;
	logD = logA-logI;
	l = logT[logI]*(1.0f-logD) + logT[logI+1]*(logD);

	lb = (l-lshift)*b;
	if(lb < -5.0f) return 0.0f;

	exp1 = 1.0f;
	// while(lb < -1.0f) { exp1 *= cexp; lb += 1.0f; }
	if(lb <= -1.0f) return 0;

	expA = -lb * ((float)expTSIZE-2.0f);
    expI = (int) expA;
    expD = expA-expI;
    return exp1 * (
		expT[expI]*(1.0f-expD) + expT[expI+1]*(expD)
	);
}


Module::Module(seaudioMasterCallback2 seaudioMaster, void *p_resvd1) : SEModule_base(seaudioMaster, p_resvd1),
	phase(0.f),
	FATOSC_TABLESIZE_MINUS_1(FATOSC_TABLESIZE-1),
	FATOSC_TABLESIZE_MINUS_1_INT(FATOSC_TABLESIZE-1),
	log16(log(16.0f)),
	cexp(exp(-1.0f))
{

}


void Module::SetupLookupTable()
{
	bool need_initialise;
	need_initialise = CreateSharedLookup( "fatosc fatness LUT", &f_table, -1, FATOSC_TABLESIZE+2 );
	need_initialise = CreateSharedLookup( "fatosc positive freq LUT", &p_table, -1, FATOSC_TABLESIZE+2 );
	need_initialise = CreateSharedLookup( "fatosc negative freq LUT", &n_table, -1, FATOSC_TABLESIZE+2 );
	need_initialise = CreateSharedLookup( "fatosc helper", &tx, -1, 16+2 );
	need_initialise = CreateSharedLookup( "fatosc log LUT", &logT, -1, logTSIZE+2 );
	need_initialise = CreateSharedLookup( "fatosc exp LUT", &expT, -1, expTSIZE+2 );
}

bool Module::CreateSharedLookup( char *p_table_name, void *p_table_pointer, float p_sample_rate, int p_size)
{
	return 0 != CallHost(seaudioMasterCreateSharedLookup, (long)p_table_pointer, p_size, p_table_name, p_sample_rate);
}

void Module::FillLookupTable()
{
	float sr = getSampleRate();

	// *** initialize the LUTs
	for(int i = 0; i < FATOSC_TABLESIZE; i++) {
		float f; // fatness index (0..1)
		float f_add, f_sub, f_range; // for fatness interpolation
		float p_v; // positive voltage
		float n_v; // negative voltage
		p_v = ((float) i / FATOSC_TABLESIZE);
		n_v = -1.f + ((float) i / FATOSC_TABLESIZE);
		f = p_v * 10.f;
		if(f <= 2.0f) {
			f_add = 0.0f;
			f_sub = 0.0f;
			f_range = 2.0f/0.25f;
		} else if(f <= 5.0f) {
			f_add = 0.25f;
			f_sub = 2.0f;
			f_range = 3.0f/0.75f;
		} else if(f <= 8.0f) {
			f_add = 1.0f;
			f_sub = 5.0f;
			f_range = 3.0f/2.0f;
		} else {
			f_add = 3.0f;
			f_sub = 8.0f;
			f_range = 2.0f/97.0f;
		}
		f_table[i] = f_add + ( (f - f_sub) / f_range );
		p_table[i] = 440.f * pow(2.f, (p_v * 10.f) - 5.f) / sr * 4.0f;
		n_table[i] = 440.f * pow(2.f, (n_v * 10.f) - 5.f) / sr * 4.0f;
		// printf("[%d] F=(%.3f) -> %.9f P=%.3f N=(%.3f) -> %.9f\n", i, p_v, f_table[i], p_table[i], n_v, n_table[i]);
	}
	// extrapolate last entries
	float a = f_table[1023];
	float b = f_table[1022];
	f_table[1024] = a + a - b;
	a = p_table[1023];
	b = p_table[1022];
	p_table[1024] = a + a - b;
	a = n_table[1023];
	b = n_table[1022];
	n_table[1024] = a + a - b;

	tx[ 0] =  1.f; tx[ 1] =  1.f; tx[ 2] = -1.f; tx[ 3] = -1.f;
	tx[ 4] = -1.f; tx[ 5] = -1.f; tx[ 6] =  1.f; tx[ 7] =  1.f;
	tx[ 8] =  1.f; tx[ 9] =  0.f; tx[10] =  1.f; tx[11] =  0.f;
	tx[12] = -1.f; tx[13] = +1.f; tx[14] = -1.f; tx[15] = +1.f;

	// *** fill the log/exp LUTs
	float v, e;

	for(int i = 0; i < logTSIZE; i++) {
		v = (float) i * 1.0 / ((float) logTSIZE-1.0f);
		logT[i] = log(sqrt(sqrt(v)));
		// printf("logT[%d]=%.6f\n", i, logT[i]);
	}
	logT[0] = logT[1]-1.0f;
	logT[logTSIZE] = log(1.0f);

	for(int i = 0; i < expTSIZE; i++) {
		v = (float) i * -1.0 / ((float) expTSIZE-1.0f);
		e = exp(v);
		expT[i] = e*e*e*e;
		// printf("expT[%d]=%.6f\n", i, expT[i]);
	}
	expT[logTSIZE] = exp(1.0f)*exp(1.0f)*exp(1.0f)*exp(1.0f);

}

void Module::open()
{

	// phase = 0.0f;
	SetupLookupTable();
	FillLookupTable();
	SET_PROCESS_FUNC(Module::sub_process);

	// call the base class
	SEModule_base::open();

}

Module::~Module()
{
	// This is where you free any memory/resources your module has created
}

// describe your module
bool Module::getModuleProperties (SEModuleProperties* properties)
{
	// describe the plugin, this is the name the end-user will see.
	properties->name = "fatosc";								// !!TODO!!

	// return a unique string 32 characters max
	// if posible include manufacturer and plugin identity
	// this is used internally by SE to identify the plug.
	// No two plugs may have the same id.
	properties->id = "dada.fatosc";

	// Info, may include Author, Web page whatever
	properties->about = "by dada";
	return true;
}

// describe the pins (plugs and parameters)
bool Module::getPinProperties (long index, SEPinProperties* properties)
{
	switch( index )								// !!TODO!! list your in / out plugs
	{
	case 0: // pitch
		properties->name				= "Pitch";
		properties->variable_address	= &input1_buffer;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "0";
		break;
	case 1: // fatness
		properties->name				= "Fatness";
		properties->variable_address	= &input2_buffer;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "0";
		break;
	case 2: // output
		properties->name				= "Signal Output";
		properties->variable_address	= &output1_buffer;
		properties->direction			= DR_OUT;
		properties->datatype			= DT_FSAMPLE;
		break;
	default:
		return false; // host will ask for plugs 0,1,2,3 etc. return false to signal when done
	};

	return true;
}

// An input plug has changed value
void Module::OnPlugStateChange(SEPin *pin)
{
	// 'transmit' new output status to next module 'downstream'
	getPin(PN_OUTPUT1)->TransmitStatusChange( SampleClock(), ST_RUN);

}
